(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2007-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(*  This example is based on an example designed for OCamlP3l.         *)
(*                                                                     *)
(***********************************************************************)

(* $Id: gnuplot.ml,v 1.12 2014/01/21 09:21:18 fclement Exp $ *)

open Config;;
open Rdec2;;

(* The gnuplot version number is set by HTMLC from the knowledge got by an
   external shell script. *)
let gnuplot_version_number = 4.0;;

let check_graphical_output = function
  | No -> ()
  | Window | Postscript as go ->
    if gnuplot_version_number <= 0. then begin
      prerr_endline
        (Printf.sprintf
           "Graphical output %s needs Gnuplot."
           (string_of_graphical_output go));
      exit 4 end
;;

(* The syntax of some gnuplot commands depends on the version number *)
let unset, pointsize =
  let us, ps =
    if gnuplot_version_number >= 4.0 then
      (fun s -> Printf.sprintf "\nunset %s" s),
      (fun () -> Printf.sprintf "")
    else
      (fun s -> Printf.sprintf "\nset no%s" s),
      (fun () -> Printf.sprintf "pointsize 0.5") in
  (fun () -> us),
  (fun () -> ps)
;;

(* Generating the gnuplot command *)

let make_isosample () (nsbdx, nsbdy) =
  Printf.sprintf "set isosample %d, %d" (nsbdx + 1) (nsbdy + 1)

and make_term_or_title () (output, res, iter) =
  match output with
  | Postscript ->
      Printf.sprintf
        "%a%a%a%a\
        \nset terminal postscript eps color solid\
        \nset output '%s%d.eps'"
        unset "xtics"
        unset "ytics"
        unset "ztics"
        unset "border"
        res iter
  | Window | No -> Printf.sprintf "\nset title 'iter = %d'" iter

and make_splot () (res, nsbd) =
  let s = ref "\nsplot [0:1][0:1][0:0.07]" in
  let to_do n =
    let resn = Printf.sprintf "%s%d" res n in
    if not (Sys.file_exists resn) then
      begin Printf.eprintf "File %s not found!\n" resn; flush stderr end;
    let file = Printf.sprintf " '%s'" resn in
    let style =
      Printf.sprintf " notitle with points pointtype 7%a," pointsize () in
    s := !s ^ file ^ style in
  for_all_subdomains nsbd to_do;
  !s

and make_reference_surface () output =
  let f, wl = "x*(1-x)*y*(1-y)", "with lines" in
  match output with
  | Postscript -> Printf.sprintf " %s notitle %s linetype 9 linewidth 4\n" f wl
  | Window | No -> Printf.sprintf " %s %s\n" f wl
;;

let make_plot_command output res iter nsbd =
  Printf.sprintf "%a%a%a%a"
    make_isosample nsbd
    make_term_or_title (output, res, iter)
    make_splot (res, nsbd)
    make_reference_surface output
;;
