(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2007-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: options.ml,v 1.11 2014/01/21 09:21:18 fclement Exp $ *)

type running_mode = Std | Tri;;

type t =
  {
   number_of_workers : int;
   vector_file : string;
   matrix_file : string;
   running_mode : running_mode;
  }
;;

let { Uaref.get = get_number_of_workers; set = set_number_of_workers; } =
  Uaref.make_aref_gen
    (Uaref.with_int_bounds (Some 1) None
      (Uaref.with_name "number_of_workers"
        (Uaref.with_single_assignment
          (Uaref.aref_spec 2))))

and { Uaref.get = get_vector_file; set = set_vector_file; } =
  Uaref.make_aref_single "vector_file" "data/demo_vec"

and { Uaref.get = get_matrix_file; set = set_matrix_file; } =
  Uaref.make_aref_single "matrix_file" "data/demo_mat"

and { Uaref.get = get_running_mode; set = set_running_mode; } =
  Uaref.make_aref_single "running_mode" "std"
;;

let mode_of_string = function
| "std" -> Std
| "tri" -> Tri
| _ -> assert false
;;

let add_options () =
  let access_ok f s =
    try Unix.access s [ Unix.R_OK; ]; f s
    with Unix.Unix_error _ ->
      Printf.eprintf "Cannot access %S for reading.\n" s;
      exit 2 in
  
  Ucommand_option.user_add_all
    (List.map
       (fun rco ->
         Ucommand_option.with_auto_short_key
           (Ucommand_option.of_regular rco))
       [
        {
         Ucommand_option.key  = "--mode";
         spec = Arg.Symbol ([ "std"; "tri"; ], set_running_mode);
         doc =
           Printf.sprintf
             "<mode>  Set the running mode of the demo.\
             \n\tDefault is %s."
             (get_running_mode ());
        };
     
        {
         Ucommand_option.key  = "--number-of-workers";
         spec = Arg.Int set_number_of_workers;
         doc =
           Printf.sprintf
             "<int>  Set the number of effective workers.\
             \n\tDefault is %d."
             (get_number_of_workers ());
        };
     
        {
         Ucommand_option.key  = "--vector-file";
         spec = Arg.String (access_ok set_vector_file);
         doc =
           Printf.sprintf
             "<file>  Set the file containing the input vector.\
             \n\tDefault is %S."
             (get_vector_file ());
        };
        
        {
         Ucommand_option.key  = "--matrix-file";
         spec = Arg.String (access_ok set_matrix_file);
         doc =
           Printf.sprintf
             "<file> Sets the file containing the input matrix.\
             \n\tDefault is %S."
             (get_matrix_file ());
        };
       ]);
;;

Ucommand_option.set_usage_msg
  (Printf.sprintf "\n%s -- Matrice-vector product" Sys.argv.(0))
;;

let get_options () =
  {
   number_of_workers = get_number_of_workers ();
   vector_file = get_vector_file ();
   matrix_file = get_matrix_file ();
   running_mode = mode_of_string (get_running_mode ());
  }
;;
