(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2007-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: IO_sockets.ml,v 1.14 2014/01/21 09:21:15 fclement Exp $ *)

(* Settings *)

let io_buff_default_size = ref 16384;;

(* Types & exceptions *)

type 'a bufferized = {
  mutable buff : string;
  mutable cur : int;
  mutable len : int;
  chan : 'a;
};;

type socket = {
  fd : Unix.file_descr;
  addr : Unix.sockaddr;
};;

type fd =
  | Socket of socket
  | Fd of Unix.file_descr
;;

type ic = fd bufferized;;
type oc = fd;;

exception IOError of string;;

(* Helpers *)

let rec intr_loop f =
  try f ()
  with Unix.Unix_error (Unix.EINTR, _, _) -> intr_loop f
;;

let ioerr msg =
  raise (IOError msg)
;;

let dump_sockaddr () addr =
  match addr with
  | Unix.ADDR_UNIX addr -> addr
  | Unix.ADDR_INET (addr, port) ->
    Printf.sprintf "%s:%d" (Unix.string_of_inet_addr addr) port
;;

let print_sockaddr oc addr =
  Printf.fprintf oc "%s" (dump_sockaddr () addr)
;;

let read_sockaddr str =
  let sb = Scanf.Scanning.from_string str in
  try
    let (addr, port) =
      Scanf.bscanf sb "%s@:%d"
        (fun a p ->
          (Unix.inet_addr_of_string a, p)) in
    Unix.ADDR_INET (addr, port)
  with _ -> failwith "Invalid address format, use www.xxx.yyy.zzz:port format"
;;

(* Bufferized data manipulation *)

let bufferize chan = {
  buff = String.create !io_buff_default_size;
  cur  = 0;
  len  = 0;
  chan = chan;
};;

let realloc_string s sz =
  let len = String.length s in
  let s' = String.create sz in
  String.blit s 0 s' 0 (min len sz);
  s'
;;

let buff_consume_data buf feeder size =
  let { cur = c; len = l; _ } = buf in
  let sz =
    let oldsz = String.length buf.buff in
    if size > oldsz
    then (buf.buff <- realloc_string buf.buff size; size)
    else oldsz
  in
  if size <= l then
    let output = String.sub buf.buff c size in
    buf.cur <- c + size;
    buf.len <- l - size;
    output
  else
    if size < sz - c then (* <= ? *)
      let output = String.create size in
      let n = feeder buf (size - l) in
      String.blit buf.buff c output 0 size;
      buf.cur <- c + size;
      buf.len <- l - size + n;
      output
    else
      let output = String.create size in
      String.blit buf.buff c buf.buff 0 l;
      buf.cur <- 0;
      let n = feeder buf (size - l) in
      String.blit buf.buff 0 output 0 size;
      buf.cur <- size;
      buf.len <- l - size + n;
      output
;;

(* Channel manipulations *)

(* (* Unused. *)
let get_addr ch =
  match ch with
  | Socket { addr = a; _ } -> a
  | Fd _ -> raise Not_found
;;
*)

let get_descr ch =
  match ch with
  | Socket { fd = descr; _ } -> descr
  | Fd descr -> descr
;;

let sock_feeder buf min_sz =
  let buf_sz = String.length buf.buff in
  let max_sz = buf_sz - (buf.cur + buf.len) in
  assert (max_sz >= min_sz);
  let descr = get_descr buf.chan in
  try
    let rec loop remaining =
      if remaining <= (max_sz - min_sz)
      then max_sz - remaining
      else let ofs = buf.cur + buf.len in
      let nread =
        intr_loop (fun () -> Unix.read descr buf.buff ofs remaining) in
      if nread = 0 then raise End_of_file else ();
      buf.len <- buf.len + nread;
      loop (remaining - nread)
    in
    loop max_sz
  with Unix.Unix_error (e, _, _) ->
    ioerr ("IO.sock_feeder: " ^ (Unix.error_message e))
;;

let send_to_channel ch str ofs n =
  let descr = get_descr ch in
  try
    let rec loop ofs len =
      let written = Unix.write descr str ofs len in
      if written = len
      then ()
      else loop (ofs + written) (len - written)
    in
    loop ofs n
  with Unix.Unix_error (e, _, _) ->
    ioerr ("IO.send_to_channel: " ^ (Unix.error_message e))
;;

let send_string ch str =
  send_to_channel ch str 0 (String.length str)
;;

let receive_of_channel ch n =
  buff_consume_data ch sock_feeder n
;;

let close_in ch =
  Unix.close (get_descr ch.chan)
;;

let close_out ch =
  Unix.close (get_descr ch)
;;

let to_channel ch x _ =
  let s = Marshal.to_string x [] in
  send_string ch s
;;

let of_channel ch =
  try
    let hs = Marshal.header_size in
    let header = receive_of_channel ch hs in
    let sz = Marshal.data_size header 0 in
    let data = receive_of_channel ch sz in
    Marshal.from_string (header ^ data) 0
  with Failure msg ->
    ioerr ("IO.of_channel: " ^ msg)
;;

let in_channel_of_ic ch =
  Unix.in_channel_of_descr (get_descr ch.chan)
;;
let ic_of_in_channel ic =
  let fd = Unix.descr_of_in_channel ic in
  bufferize (Fd fd)
;;

let out_channel_of_oc ch =
  Unix.out_channel_of_descr (get_descr ch)
;;
let oc_of_out_channel oc =
  let fd = Unix.descr_of_out_channel oc in
  Fd fd
;;

let channels_of_sock sock =
  let socket =
    Socket { fd = sock; addr = Unix.getsockname sock; }
  in
  (bufferize socket, socket)
;;

let stdin = bufferize (Fd Unix.stdin);;
let stdout = Fd Unix.stdout;;

(* Unix sockets helpers *)

let create_socket () =
  Unix.socket Unix.PF_INET Unix.SOCK_STREAM 0
;;

let get_port addr =
  match addr with
  | Unix.ADDR_INET (_, port) -> port
  | Unix.ADDR_UNIX _ -> raise Not_found
;;

let get_addr addr =
  match addr with
  | Unix.ADDR_INET (a, _) -> a
  | Unix.ADDR_UNIX _ -> raise Not_found
;;

let connect addr =
  let fd = create_socket () in
  try
    Unix.connect fd addr;
    channels_of_sock fd
  with Unix.Unix_error (e, _, _) ->
    Unix.close fd;
    ioerr ("IO.connect: " ^ (Unix.error_message e))
;;

let connect_fd addr =
  let fd = create_socket () in
  try
    Unix.connect fd addr;
    fd
  with Unix.Unix_error (e, _, _) ->
    Unix.close fd;
    ioerr ("IO.connect_fd: " ^ (Unix.error_message e))
;;

let create_bound_socket port =
  let port =
    match port with
    | Some p -> p
    | None -> 0 in
  let sock = create_socket () in
  try
    Unix.bind sock (Unix.ADDR_INET (Unix.inet_addr_any, port));
    Unix.listen sock 1;
    (sock, Unix.getsockname sock)
  with Unix.Unix_error (e, _, _) ->
    Unix.close sock;
    ioerr ("IO.create_bound_socket: " ^ (Unix.error_message e))
;;

let wait_connection sock closeb =
  let fd, addr = Unix.accept sock in
  if closeb then Unix.close sock else ();
  channels_of_sock fd, (get_addr addr)
;;

(* Process management *)

type pid = int;;

type fork_data =
   | Child
   | Parent of pid
;;

let fork () =
  try
    flush stderr; (* Avoid duplicating output *)
    match Unix.fork () with
    | 0 -> Child
    | pid -> Parent pid
  with Unix.Unix_error (e, _, _) ->
    ioerr ("IO.fork: " ^ (Unix.error_message e))
;;

let open_process_in s =
  try ic_of_in_channel (Unix.open_process_in s)
  with Unix.Unix_error (e, _, _) ->
    ioerr ("IO.open_process_in: " ^ (Unix.error_message e))
;;

let open_process s =
  try
    let ic, oc = Unix.open_process s in
    (ic_of_in_channel ic,
     oc_of_out_channel oc)
  with Unix.Unix_error (e, _, _) ->
    ioerr ("IO.open_process: " ^ (Unix.error_message e))
;;

(* Files *)

let open_in_out str =
  try
    let fd = intr_loop
      (fun () -> Unix.openfile str [ Unix.O_RDWR; Unix.O_SYNC; ] 0) in
    (bufferize (Fd fd), Fd fd)
  with Unix.Unix_error (e, _, _) ->
    ioerr ("IO.open_in_out: " ^ (Unix.error_message e))
;;

let open_in str =
  try
    let fd = intr_loop
      (fun () -> Unix.openfile str [ Unix.O_RDONLY; Unix.O_SYNC; ] 0) in
    bufferize (Fd fd)
  with Unix.Unix_error (e, _, _) ->
    ioerr ("IO.open_in: " ^ (Unix.error_message e))
;;

let open_default_prems = 0o640;;

let open_out str =
  try
    let fd = intr_loop
      (fun () -> Unix.openfile str [ Unix.O_WRONLY; Unix.O_SYNC; ]
        open_default_prems) in
    Fd fd
  with Unix.Unix_error (e, _, _) ->
    ioerr ("IO.open_out: " ^ (Unix.error_message e))
;;

(* Fifo and pipes *)

let pipe () =
  try
    let (ic, oc) = Unix.pipe () in
    (bufferize (Fd ic), Fd oc)
  with Unix.Unix_error (e, _, _) ->
    ioerr ("IO.safe_pipe: " ^ (Unix.error_message e))
;;

let fifo_default_perms = 0o640;;

let mk_fifo k str =
  try
    intr_loop (fun () -> Unix.mkfifo str fifo_default_perms);
    k str
  with Unix.Unix_error (e, _, _) ->
    ioerr ("IO.openfifo: " ^ (Unix.error_message e))
;;

let openfifo_in str =
  mk_fifo open_in str
;;

let openfifo_out str =
  mk_fifo open_out str
;;
