(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2007-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: main.ml,v 1.30 2014/01/21 09:21:15 fclement Exp $ *)

(* This is the sklml compiler *)

open Sklml_install_data;;

type semantics_mode = Seq | Par;;

type compilation_mode = Byt | Bin;;

type compilation_options = {
  compilation_mode : compilation_mode;
  semantics_mode : semantics_mode;
  lib_root_dir : string;
  pp_cmd : string;
  ocamlc_options : string list;
};;

let string_of_semantics_mode = function
  | Seq -> "seq"
  | Par -> "par"
;;

let dir_of_semantics_mode dir = function
  | Seq -> mk_lib_path dir ["sequential"; "lib"]
  | Par -> mk_lib_path dir ["parallel"; "lib"]
;;

let get_compiler_and_libs compilation_mode opts =
  let (compiler, lib_suffix) =
    match compilation_mode with
    | Bin -> (ocaml_compiler.bin, ".cmxa")
    | Byt -> (ocaml_compiler.byt, ".cma") in
  let libs =
    if List.exists (( = ) "-c") opts then [] else
    List.map
      (fun x -> x ^ lib_suffix)
      [ "bigarray"; "unix"; "ulib"; "lib_pio";
        "parlib"; "sklml"; "sklml_extra"; ] in
  (compiler, libs)
;;

let get_include_dirs dir mode =
  [
   mk_lib_path dir [ "private"; "lib"; "ocaml"; "ulib"; ];
   mk_lib_path dir [ "private"; "lib"; "ocaml"; "pio"; ];
   mk_lib_path dir [ "parallel"; "parlib"; ];
   mk_lib_path dir [ "common"; ];
   dir_of_semantics_mode dir mode;
   mk_lib_path dir [ "extra"; ];
  ]
;;

let unwords = String.concat " ";;

let mk_command
    {
     compilation_mode = comp_mode;
     semantics_mode = sem_mode;
     lib_root_dir = dir;
     pp_cmd = pp_cmd;
     ocamlc_options = caml_options;
    } =
  let (compiler, sklml_libs) = get_compiler_and_libs comp_mode caml_options in
  let include_dirs = get_include_dirs dir sem_mode in
  let include_options =
    List.flatten (List.map (fun d -> [ "-I"; d; ]) include_dirs) in
  let pp_option = [ "-pp"; Printf.sprintf "'%s'" pp_cmd; ] in
  let generic_options = include_options @ pp_option in
  unwords ([compiler] @ generic_options @ sklml_libs @ caml_options)
;;

let fail msg = Printf.eprintf "sklmlc: %s\n" msg; exit 1;;

let parse_argv args =
  let (get_annons, accum_annons) =
    let annons = ref [] in
    (fun () -> !annons),
    (fun px s -> annons := !annons @ [px ^ s]) in
  let (get_lib_root_dir, set_lib_root_dir) =
    let dir = ref default_lib_root_dir in
    (fun () -> !dir),
    (fun s -> dir := s)
  and (get_semantics_mode, set_semantics_mode) =
    let m = ref Seq in
    (fun () -> !m),
    (fun s -> match s with
      | "seq" -> m := Seq
      | "par" -> m := Par
      | _ -> assert false)
  and (get_debug, set_debug) =
    let d = ref false in
    (fun () -> !d),
    (fun () -> d := true)
  and (get_compilation_mode, set_compilation_mode_to_native_code) =
    let m = ref Byt in
    (fun () -> !m),
    (fun () -> m := Bin)
  in
  let speclist =
    [
     "-mode", Arg.Symbol (["seq"; "par"], set_semantics_mode),
     (Printf.sprintf
        "  Set the semantics mode. Default is %S."
        (string_of_semantics_mode (get_semantics_mode ())));
     
     "-O", Arg.Unit set_compilation_mode_to_native_code,
     " Invoke native-code compilation. \
       Default is to invoke bytecode compilation.";
     
     "-sklml-lib-root-dir", Arg.String set_lib_root_dir,
     "<dir>  Set the library root directory of the Sklml system.";
     
     "-ext", Arg.String (accum_annons "-"),
     "<arg>  Internal use only.";
     
     "-debug", Arg.Unit set_debug,
     " Turn the debugging mode on. Default is off.";
     
     "-where",
     Arg.Unit (fun () -> prerr_endline (get_lib_root_dir ()); exit 0),
     " Print the location of Sklml system and exit.";
     
     "-v",
     Arg.Unit
       (Upph_id_card.make_print_short_version Sklml_version.pph_id_card),
     " Print the short form of the version information and exit.";
     
     "-version",
     Arg.Unit
       (Upph_id_card.make_print_long_version Sklml_version.pph_id_card),
     " Print the long form of the version information and exit.";
    ] in
  let keys =
    [ "-help"; "--help"; ] @ List.map (fun (x, _, _) -> x) speclist in
  let args = Array.of_list
    (Array.fold_left (fun l s ->
      if not (List.mem s keys) then
        let key = String.sub s 1 (String.length s - 1) in
        if s.[0] = '-'
        then l @ [ "-ext"; key ]
        else l @ [s]
      else l @ [s]) [] args) in
  let head_usage =
    "Usage: sklmlc [OPTIONS] FILES...\n\n\
     Options are (in addition to those of the OCaml compilers):" in
  try
    Arg.parse_argv args speclist (accum_annons "") head_usage;
    let lib_root_dir = get_lib_root_dir () in
    {
     compilation_mode = get_compilation_mode ();
     semantics_mode = get_semantics_mode ();
     lib_root_dir = lib_root_dir;
     pp_cmd = mk_pp_cmd (mk_camlp45_pa_skl lib_root_dir);
     ocamlc_options = get_annons ();
    },
    get_debug ()
  with
  | Arg.Bad _ -> fail "invalid usage, see help using -help option"
  | Arg.Help m -> print_string m; exit 0
  | Failure m -> fail ("invalid usage, " ^ m)
;;

let dprintf b =
  (if b then Printf.fprintf else Printf.ifprintf) stdout
;;

let main () =
  let (compile_opts, dbg) = parse_argv Sys.argv in
  let compile_cmd = mk_command compile_opts in
  dprintf dbg "will run : %s\n" compile_cmd;
  let ret = Sys.command compile_cmd in
  dprintf dbg
    (if ret = 0
     then format_of_string "compilation finished\n"
     else format_of_string "there were errors\n");
  exit ret
;;

let _ = main ();;
